#include <stdio.h>
#include <iostream>
#include "i3d/image3d.h"
#include "i3d/vector3d.h"

using namespace std;
using namespace i3d;


/**
 * Funkce ocekava oba vstupni obrazy \e in a \e pattern, ze jsou jiz defacto masky.
 * Maske \e pattern je postupne prikladana na vstupni obraz \e in a pro kazde takove
 * prilozeni se pocita velikost spolecneho pruniku obou masek (tedy, pod kolika pixely
 * z masky \e pattern se nachazely take pixely ze vstupu \e in). Vysledne pocty jsou
 * ukladany do vystupniho obrazu \e out.
 */
void PatternMatching(Image3d<GRAY8> const &in,
							Image3d<GRAY8> const &pattern,
							Image3d<GRAY8> &out,
							const int max_count)
{
	//zajisti, ze alokaci vystupniho obrazu
	out.CopyMetaData(in);

	//time-saver :-)
	const i3d::Vector3d<size_t> centre=pattern.GetSize()/(size_t)2;

	//pro kazdy pixel vstupniho obrazu projdi vzdy cely vzor a spocitej kolik je tam shod
	for (size_t y=0; y < in.GetSizeY()-pattern.GetSizeY(); ++y)
		for (size_t x=0; x < in.GetSizeX()-pattern.GetSizeX(); ++x) {
			int matches=0;
			for (size_t yy=0; yy < pattern.GetSizeY(); ++yy)
				for (size_t xx=0; xx < pattern.GetSizeX(); ++xx)
					matches+=(in.GetVoxel(x+xx,y+yy,0) && pattern.GetVoxel(xx,yy,0))?  1 : 0;
			out.SetVoxel(i3d::Vector3d<size_t>(x,y,0)+centre,100*matches/max_count);
		}
}


int main(int argc,char **argv) {

	//nacteni vstupniho obrazu
	Image3d<GRAY8> img;
	//img.ReadImageBMP("birds_mask170.bmp");
	img.ReadImageBMP("birds_mask183.bmp");

	//nacteni vzoru
	Image3d<GRAY8> p1,p2,p3;
	p1.ReadImageBMP("p1.bmp");
	p2.ReadImageBMP("p2.bmp");
	p3.ReadImageBMP("p3.bmp");

	//vysledky aplikace vzoru
	Image3d<GRAY8> out1,out2,out3;

	//postupne aplikovani vzoru
	int pattern_size=0;
	for (size_t i=0; i < p1.GetImageSize(); ++i)
		if (p1.GetVoxel(i)) ++pattern_size;
	PatternMatching(img,p1,out1,pattern_size);
	cout << "pattern matching #1 hotovy...\n";

	pattern_size=0;
	for (size_t i=0; i < p2.GetImageSize(); ++i)
		if (p2.GetVoxel(i)) ++pattern_size;
	PatternMatching(img,p2,out2,pattern_size);
	cout << "pattern matching #2 hotovy...\n";

	pattern_size=0;
	for (size_t i=0; i < p3.GetImageSize(); ++i)
		if (p3.GetVoxel(i)) ++pattern_size;
	PatternMatching(img,p3,out3,pattern_size);
	cout << "pattern matching #3 hotovy...\n";

	//ulozeni mezivysledku
	out1.SaveImageBMP("matches1.bmp");
	out2.SaveImageBMP("matches2.bmp");
	out3.SaveImageBMP("matches3.bmp");

	//nalezeni pixel-wise maxima: kazdy pixel bude mit hodnotu 0-100,
	//ktera ukazuje, jak nejlepe (=100) na tento pixel pasoval nejaky
	//z pouzitych vzoru
	for (size_t i=0; i < out1.GetImageSize(); ++i)
		out1.SetVoxel(i,std::max( out1.GetVoxel(i),
			std::max(out2.GetVoxel(i), out3.GetVoxel(i)) ));

	//ulozeni mezivysledku
	out1.SaveImageBMP("matches_combined.bmp");


	//do obrazu 'img' budeme ukladat komponenty,
	//ktere maji reprezentovat detekovane ptaky
	img.GetVoxelData()=0; //vymazani obsahu
	//img=VISITED obrazek

	//sem se vzdy zakresli prvni pixel od kazde nove nalezene
	//komponenty, to bude pravdepodobne reprezentovat zhruba stredy
	//nalezenych ptaku
	Image3d<GRAY8> ptaci;
	ptaci.CopyMetaData(out1); //take smaze obsah
	//ptaci=ZARODKY

	//pokud se zachyti novy potencialni zarodek, bude se
	//skutecne povazovat za zarodek dalsiho ptaka jenom pokud je alespon
	//'radius' pixelu daleko od jakekoliv jiz objevene komponenty
	const int radius=4;

	//pocet nalezenych ptaku
	int pocet_ptaku=0;


	//pres vsechny prahy a vsechny pixely s danym prahem
	for (GRAY8 prah=100; prah > 66; --prah)
		for (size_t i=0; i < out1.GetImageSize(); ++i)
			if (out1.GetVoxel(i) == prah)
			{
				//zkontrolujeme, zda neni pobliz jiz jiny drive nalezeny
				if (img.GetVoxel(i) == 0)
				{
					//pobliz neni zadny drive objeveny region, nasel jsem zarodek ptaka
					++pocet_ptaku;
					ptaci.SetVoxel(i,200); //reprezentant nalezeneho ptaka
					img.SetVoxel(i,200);   //jeste vyznaci zarodek
				}

				//tak jako tak, vyznacime region "blizke obsazenosti" 
				const int X=(int)img.GetX(i);
				const int Y=(int)img.GetY(i);
				for (int y=-radius; y <= radius; ++y)
					for (int x=-radius; x <= radius; ++x)
						//if (img.Include(X+x,Y+y,0)) -- neni nutne, protoze mame okraje s intenzitou 0
							img.SetVoxel(X+x,Y+y,0,
								//100                                 //natvrdo "vypali" znacku obsazenosti
								std::max(100,(int)img.GetVoxel(X+x,Y+y,0)) //zachova vyznacene zarodky
								);

				//debug:
				char fn[50];
				sprintf(fn,"mezivysledek%03d.bmp",prah);
				img.SaveImageBMP(fn);
			}

	ptaci.SaveImageBMP("zarodky.bmp");
	cout << "nasel jsem nakonec " << pocet_ptaku << " ptaku\n";

	return(0);
}
